#include<ylm.h>

/*! \fn complex Ylm(long l, long m, double theta, double phi)
    evaluation of Ylm from Racah
    Expansion yields polynomial in sin(theta)
    using 5.2.(17) from Varshalovich et al. (1988).
*/

complex ylm(long l, long m, double theta, double phi)
{
  complex ylm;
  long L, s;
  double fact;

  if (l<0) {fprintf(stderr, "err: Ylm: l<0\n"); exit(-69);}
  if (l<fabs(m)) {fprintf(stderr, "err: Ylm: l<|m|\n"); exit(-70);}
      
  if  (!((l-m)&1)) {L=l; fact=1.0;} else {L=l-1; fact=cos(theta);};

  if (sin(theta) != 0.0)
  {
    for(s=(long)(fabs(m)); s<=L; s=s+2)
    {
       ylm = ylm + pow(-1.0,(s+m)/2.0) * factorial(l+s) / 
	    (factorial((s+m)/2)*factorial((s-m)/2)*pow(2.0,double(s))) *
            (factorial((L+m)/2)*factorial((L-m)/2)) / 
	    (factorial((L+s)/2)*factorial((L-s)/2)) *
             pow(sin(theta),double(s-fabs(m)));
    }

    ylm = ylm * sqrt((l+l+1)/(4*M_PI*factorial(l+m)*factorial(l-m))) *  
    pow(sin(theta),double(fabs(m))) * fact * exp(complex(0.0, m*phi));
  }
  else if (m == 0) 
  {
    ylm = sqrt((l+l+1)/M_PI) * fact / 2;
  } 
  else 
  {
    ylm = 0.0; 
  }

  return(ylm);
}
