#include<iostream.h>
#include<stdio.h>

#include<wavefunction.h>
#include<fluid.h>
#include<grid.h>
#include<hamop.h>
#include<winop.h>
#include<xml_parse.h>
#include<ylm.h>


//
// definitions of potentials, nuclear_charge etc.
//
#include<../hydrogen/potentials.cc>

int main(int argc, char **argv)
{
  int  iv;

  long ngps_x, ngps_y, ngps_z, dimens, index, i, no_energ, no_theta, no_phi,
       lnoofwfoutput, ltheta, lphi, lenergy, no_dir, 
       m_limit, m_index, l_index, ylm_index;

  double delt_x, delta_energy, energy_init, energ, gamma, w;

  complex summingup, result_tot;

  FILE *file_wf, *file_res, *file_info;

  char cstr_fname_res[1000], cstr_dname_wf[1000],
       cstr_fname_info[1000], cstr_fname_wf[1000];

  hamop  hamilton;

  fluid  theta, phi, V_ee_0;
  grid  g_angle, g_load, g;
  wavefunction fullchi, angle_result, result_lsub, ylm_array,
    staticpot, wf, wf_load;

  //
  //  set a few essential parameters for the calculation of the spectra
  //
  lnoofwfoutput       = 1;      // 0 -- initial state, 1 -- final state.
  no_energ            = 500;   // number of energy bins
  gamma               = 1.0e-3; // half width of the window operator (72)
  energy_init         = -0.55;  // first energy; last energy then is
                                // energy_init + (no_energ-1)*2*gamma

  // define for how many angles (77) should be evaluated 
  no_theta            = 1; // no_theta angles between 0 and (1-1/no_theta)*pi   
  no_phi              = 1; // no_phi angles between 0 and (1-1/no_phi)*2*pi 



  ampl_im             = 0.0; // imaginary potential is set to zero.
  no_dir              = no_theta * no_phi;
  delta_energy        = 2*gamma; // width of the bin is 2*gamma

  //
  // specify where the wavefunction (to be analyzed) is located
  // 
  sprintf(cstr_dname_wf, "../hydrogen/res/");
  sprintf(cstr_fname_wf,"%s/hydrogen_re_wf-000-0.600000-34-2000-2.00e-01.dat", cstr_dname_wf);
  //
  // get the info for the following calculation from the real-time
  // propagation output
  //
  sprintf(cstr_fname_info,"%s/hydrogen_re_info-000-0.600000-34-2000-2.00e-01.dat", cstr_dname_wf);
  file_info = fopen(cstr_fname_info, "r");  
  if (file_info==NULL) {
    fprintf(stdout, "%s can't be opened!\n", cstr_fname_info); exit(-10);}  

  xml_parse_fgetd(&nuclear_charge,file_info,"<nuclear_charge>",
    "</nuclear_charge>");
  xml_parse_fgetl(&ngps_x, file_info,"<ngps_x>", "</ngps_x>");
  xml_parse_fgetl(&ngps_y, file_info,"<ngps_y>", "</ngps_y>");
  xml_parse_fgetl(&ngps_z, file_info,"<ngps_z>", "</ngps_z>");
  xml_parse_fgetd(&delt_x,file_info,"<delt_x>","</delt_x>");
  xml_parse_fgetl(&dimens,file_info,"<dimens>","</dimens>");
  xml_parse_fgetd(&w,file_info,"<w>","</w>");
  fclose(file_info);

  sprintf(cstr_fname_res,  "./res/res-%lf-%ld-%ld-%ld-%.3le-%.3le.dat",
    w, dimens, ngps_x, ngps_y, delta_energy, gamma);

  iv=1;
  fprintf(stdout, " WINOP: Calculation of one-electron spectra\n");
  fprintf(stdout, " (C) Copyright by Bauer D and Koval P, Heidelberg (2005)\n");
  fprintf(stdout, " -------------------------------------------------------\n");

  file_wf   = fopen(cstr_fname_wf,"r");
  file_res = fopen(cstr_fname_res,"w");
 
  if (iv==1)
  { 
    fprintf(stdout, "%s will be read\n", cstr_fname_wf);
    fprintf(stdout, "%s will be written\n", cstr_fname_res);
  }

  if (file_wf==NULL) {
    fprintf(stderr, "%s can't be opened!\n", cstr_fname_wf); exit(-10);}

  // specific angles
  theta.init(no_theta);   
  for(ltheta=0; ltheta<no_theta; ltheta++)
    theta[ltheta] = ltheta * M_PI / no_theta;

  phi.init(no_phi); 
  for(lphi=0; lphi<no_phi; lphi++) 
    phi[lphi] = lphi * 2 * M_PI / no_phi;

  // this grid is used to store the Y_lm; do not touch this !!!
  g_angle.set_dim(44);   // 44 works for both propagation mode 34 and 44
  g_angle.set_ngps(no_dir, ngps_y, 0); 
  g_angle.set_delt(0.0);
  g_angle.set_offs(0,0,0);

  // this is the grid for the wavefunction that is to be analyzed
  // (data is taken from the info file)
  g_load.set_dim(dimens);
  g_load.set_ngps(ngps_x,ngps_y,ngps_z);
  g_load.set_delt(delt_x);
  g_load.set_offs(0,0,0);

  // this is the grid on which the wavefunction is analyzed
  // delt_x must be the same for g and g_load
  // increasing ngps_x improves the resolution in the continuum 
  g.set_dim(dimens);
  g.set_ngps(ngps_x,ngps_y,ngps_z); 
  g.set_delt(delt_x);
  g.set_offs(0,0,0);
  

  
  angle_result.init(no_dir);

  if (dimens==34)
  {
    fullchi.init(g.ngps_x()*g.ngps_y()); 
    result_lsub.init(g.ngps_y());
  }
else if (dimens==44)
  {
    fullchi.init(g.ngps_x()*g.ngps_y()*g.ngps_y()); 
    result_lsub.init(g.ngps_y()*g.ngps_y());
  }
  else
  { 
    fprintf(stderr, "err: wrong dimens!=34 and dimens!=44!\n"); exit(12);
  }
  
  V_ee_0.init(g.ngps_x());
  ylm_array.init(g_angle.size());

  //
  // the Hamiltonian
  //
  hamilton.init(g, vecpot_x, vecpot_y, vecpot_z, scalarpotx,
                 scalarpoty, scalarpotz, imagpot, field);

  staticpot.init(g.size());
  staticpot.calculate_staticpot(g,hamilton);

  // *** the wavefunction array 
  wf.init(g.size());
  wf_load.init(g_load.size());

  // ***** calculate all the ylm's needed
  for (l_index=0; l_index<g_angle.ngps_y(); l_index++)
  {
  if (g.dimens()==34) {m_limit = 0;} else {m_limit = l_index;}
  for (m_index=-m_limit; m_index<=m_limit; m_index++)
  {
    for (ltheta=0; ltheta<no_theta; ltheta++)
    {
    for (lphi=0; lphi<no_phi; lphi++)
    {
      ylm_index = g_angle.rlmindex(ltheta+lphi*no_theta, l_index, m_index);
      ylm_array[ylm_index]=ylm(l_index,m_index,theta[ltheta],phi[lphi]);

    }}}}

  if (iv==1) cout << "Reading ... " << endl;

  wf_load.init(g_load, file_wf, lnoofwfoutput, iv);
  fclose(file_wf);

  if (iv==1) cout << "... done." << endl;
  
  wf.regrid(g,g_load,wf_load);
  
  //
  // Stdout
  //
  if (iv==1)
  {
    fprintf(stdout, "Grid: \n", g.ngps_x());
    fprintf(stdout, "g.ngps_x() = %ld\n", g.ngps_x());
    fprintf(stdout, "g.ngps_y() = %ld\n", g.ngps_y());
    fprintf(stdout, "g.ngps_z() = %ld\n", g.ngps_z());
    fprintf(stdout, "g.dimens() = %ld\n", g.dimens());
    fprintf(stdout, "g.delt_x() = %20.15le\n", g.delt_x());
    fprintf(stdout, "nuclear_charge    = %20.15le\n", nuclear_charge);
    fprintf(stdout, "cstr_fname_wf = %s\n", cstr_fname_wf);
    fflush(stdout);
  }

  //
  // ******Starting cycle******
  //

  // *********************************
  // *** run over the energies
  // ********************************* 
  for (lenergy=0; lenergy<no_energ; lenergy++)
  {
    energ = energy_init + lenergy * delta_energy;

    fprintf(file_res, "%.15le ", energ);

    winop_fullchi( fullchi, result_lsub, &result_tot, energ, gamma,
      staticpot, V_ee_0, nuclear_charge, g, wf, iv);


    // write the partial results (i.e., for individual l (and m)) to file
    for(l_index=0; l_index<g.ngps_y(); l_index++)
      {
	if (g.dimens()==44) 
	  { 
	    m_limit=l_index;
	    for(m_index=-m_limit; m_index<=m_limit; m_index++)
	      {
		fprintf(file_res, "%.15le ", 
			real(result_lsub[((l_index+1)*(l_index+1)-(l_index+1)+m_index)]));
	      }; 
	  }
	else
	  {
	    fprintf(file_res, "%.15le ", 
		    real(result_lsub[l_index]));
	  };
      };
    // write the total result to file
    fprintf(file_res, "%.15le ", real(result_tot));

    angle_result.nullify();
    for (lphi=0; lphi<no_phi; lphi++){
    for (ltheta=0; ltheta<no_theta; ltheta++){

      if ((iv==4))
        cout << " angles theta, phi = "<<theta[ltheta]<<", "<<phi[lphi]<<endl;

      for (long rindex=0; rindex<g.ngps_x(); rindex++)
      {
	summingup=complex(0.0,0.0);
	for (l_index=0; l_index<g.ngps_y(); l_index++)
	{
          if (g.dimens()==34) {m_limit=0;} else {m_limit=l_index;}
	  for (m_index=-m_limit; m_index<=m_limit; m_index++)
	  {
	    index=g.index(rindex,l_index,m_index,0);
	    summingup+=fullchi[index]*
             ylm_array[g_angle.rlmindex(ltheta+no_theta*lphi,l_index,m_index)];
	  }
        }
	angle_result[ltheta+no_theta*lphi]+=summingup*conj(summingup);
      }
      angle_result[ltheta+no_theta*lphi]*=g.delt_x();

      fprintf(file_res, "%.15le ", real(angle_result[ltheta+no_theta*lphi]));
    }}
    fprintf(file_res, "\n");
    fflush(file_res);
    
  } // end of energy loop
  fclose(file_res);
  if (iv==1) printf("%s is written\n", cstr_fname_res);

  if (iv==1) cout << "Hasta la vista... " << endl;
}


